# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'importexportdialog.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(708, 646)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(Dialog.sizePolicy().hasHeightForWidth())
        Dialog.setSizePolicy(sizePolicy)
        self.gridLayout = QtWidgets.QGridLayout(Dialog)
        self.gridLayout.setObjectName("gridLayout")
        self.pick_sunspec_tables_c = QtWidgets.QPushButton(Dialog)
        self.pick_sunspec_tables_c.setObjectName("pick_sunspec_tables_c")
        self.gridLayout.addWidget(self.pick_sunspec_tables_c, 3, 3, 1, 1)
        self.label_2 = QtWidgets.QLabel(Dialog)
        self.label_2.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 1, 0, 1, 1)
        self.sunspec_tables_c = QtWidgets.QLineEdit(Dialog)
        self.sunspec_tables_c.setObjectName("sunspec_tables_c")
        self.gridLayout.addWidget(self.sunspec_tables_c, 3, 2, 1, 1)
        self.sil_c_label = QtWidgets.QLabel(Dialog)
        self.sil_c_label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.sil_c_label.setObjectName("sil_c_label")
        self.gridLayout.addWidget(self.sil_c_label, 8, 0, 1, 1)
        self.pick_sil_c = QtWidgets.QPushButton(Dialog)
        self.pick_sil_c.setObjectName("pick_sil_c")
        self.gridLayout.addWidget(self.pick_sil_c, 8, 3, 1, 1)
        self.hierarchy = QtWidgets.QLineEdit(Dialog)
        self.hierarchy.setObjectName("hierarchy")
        self.gridLayout.addWidget(self.hierarchy, 1, 2, 1, 1)
        self.label_3 = QtWidgets.QLabel(Dialog)
        self.label_3.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 5, 0, 1, 1)
        self.can = QtWidgets.QLineEdit(Dialog)
        self.can.setObjectName("can")
        self.gridLayout.addWidget(self.can, 0, 2, 1, 1)
        self.tables_c_label = QtWidgets.QLabel(Dialog)
        self.tables_c_label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.tables_c_label.setObjectName("tables_c_label")
        self.gridLayout.addWidget(self.tables_c_label, 2, 0, 1, 1)
        self.sil_c = QtWidgets.QLineEdit(Dialog)
        self.sil_c.setObjectName("sil_c")
        self.gridLayout.addWidget(self.sil_c, 8, 2, 1, 1)
        self.spreadsheet_user = QtWidgets.QLineEdit(Dialog)
        self.spreadsheet_user.setObjectName("spreadsheet_user")
        self.gridLayout.addWidget(self.spreadsheet_user, 6, 2, 1, 1)
        self.spreadsheet = QtWidgets.QLineEdit(Dialog)
        self.spreadsheet.setObjectName("spreadsheet")
        self.gridLayout.addWidget(self.spreadsheet, 5, 2, 1, 1)
        self.remove_smdx = QtWidgets.QPushButton(Dialog)
        self.remove_smdx.setObjectName("remove_smdx")
        self.gridLayout.addWidget(self.remove_smdx, 12, 3, 1, 1)
        self.pick_spreadsheet = QtWidgets.QPushButton(Dialog)
        self.pick_spreadsheet.setObjectName("pick_spreadsheet")
        self.gridLayout.addWidget(self.pick_spreadsheet, 5, 3, 1, 1)
        self.pick_smdx = QtWidgets.QPushButton(Dialog)
        self.pick_smdx.setObjectName("pick_smdx")
        self.gridLayout.addWidget(self.pick_smdx, 11, 3, 1, 1)
        self.pick_spreadsheet_user = QtWidgets.QPushButton(Dialog)
        self.pick_spreadsheet_user.setObjectName("pick_spreadsheet_user")
        self.gridLayout.addWidget(self.pick_spreadsheet_user, 6, 3, 1, 1)
        self.label = QtWidgets.QLabel(Dialog)
        self.label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)
        self.sunspec_tables_c_label = QtWidgets.QLabel(Dialog)
        self.sunspec_tables_c_label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.sunspec_tables_c_label.setObjectName("sunspec_tables_c_label")
        self.gridLayout.addWidget(self.sunspec_tables_c_label, 3, 0, 1, 1)
        self.tables_c = QtWidgets.QLineEdit(Dialog)
        self.tables_c.setObjectName("tables_c")
        self.gridLayout.addWidget(self.tables_c, 2, 2, 1, 1)
        self.pick_tables_c = QtWidgets.QPushButton(Dialog)
        self.pick_tables_c.setObjectName("pick_tables_c")
        self.gridLayout.addWidget(self.pick_tables_c, 2, 3, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.from_directory = QtWidgets.QPushButton(Dialog)
        self.from_directory.setObjectName("from_directory")
        self.horizontalLayout.addWidget(self.from_directory)
        self.buttons = QtWidgets.QDialogButtonBox(Dialog)
        self.buttons.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttons.setObjectName("buttons")
        self.horizontalLayout.addWidget(self.buttons)
        self.gridLayout.addLayout(self.horizontalLayout, 14, 0, 1, 4)
        self.sunspec_c = QtWidgets.QLineEdit(Dialog)
        self.sunspec_c.setObjectName("sunspec_c")
        self.gridLayout.addWidget(self.sunspec_c, 7, 2, 1, 1)
        self.smdx_list = QtWidgets.QListWidget(Dialog)
        self.smdx_list.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.smdx_list.setObjectName("smdx_list")
        self.gridLayout.addWidget(self.smdx_list, 11, 2, 3, 1)
        self.pick_hierarchy = QtWidgets.QPushButton(Dialog)
        self.pick_hierarchy.setObjectName("pick_hierarchy")
        self.gridLayout.addWidget(self.pick_hierarchy, 1, 3, 1, 1)
        self.label_4 = QtWidgets.QLabel(Dialog)
        self.label_4.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 6, 0, 1, 1)
        self.interface_c_label = QtWidgets.QLabel(Dialog)
        self.interface_c_label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.interface_c_label.setObjectName("interface_c_label")
        self.gridLayout.addWidget(self.interface_c_label, 9, 0, 1, 1)
        self.pick_interface_c = QtWidgets.QPushButton(Dialog)
        self.pick_interface_c.setObjectName("pick_interface_c")
        self.gridLayout.addWidget(self.pick_interface_c, 9, 3, 1, 1)
        self.sunspec_c_label = QtWidgets.QLabel(Dialog)
        self.sunspec_c_label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.sunspec_c_label.setObjectName("sunspec_c_label")
        self.gridLayout.addWidget(self.sunspec_c_label, 7, 0, 1, 1)
        self.interface_c = QtWidgets.QLineEdit(Dialog)
        self.interface_c.setObjectName("interface_c")
        self.gridLayout.addWidget(self.interface_c, 9, 2, 1, 1)
        self.pick_can = QtWidgets.QPushButton(Dialog)
        self.pick_can.setObjectName("pick_can")
        self.gridLayout.addWidget(self.pick_can, 0, 3, 1, 1)
        self.smdx_label = QtWidgets.QLabel(Dialog)
        self.smdx_label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.smdx_label.setObjectName("smdx_label")
        self.gridLayout.addWidget(self.smdx_label, 11, 0, 1, 1)
        self.pick_sunspec_c = QtWidgets.QPushButton(Dialog)
        self.pick_sunspec_c.setObjectName("pick_sunspec_c")
        self.gridLayout.addWidget(self.pick_sunspec_c, 7, 3, 1, 1)
        self.sunspec_bitfields_c_label = QtWidgets.QLabel(Dialog)
        self.sunspec_bitfields_c_label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.sunspec_bitfields_c_label.setObjectName("sunspec_bitfields_c_label")
        self.gridLayout.addWidget(self.sunspec_bitfields_c_label, 4, 0, 1, 1)
        self.sunspec_bitfields_c = QtWidgets.QLineEdit(Dialog)
        self.sunspec_bitfields_c.setObjectName("sunspec_bitfields_c")
        self.gridLayout.addWidget(self.sunspec_bitfields_c, 4, 2, 1, 1)
        self.pick_sunspec_bitfields_c = QtWidgets.QPushButton(Dialog)
        self.pick_sunspec_bitfields_c.setObjectName("pick_sunspec_bitfields_c")
        self.gridLayout.addWidget(self.pick_sunspec_bitfields_c, 4, 3, 1, 1)

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Dialog"))
        self.pick_sunspec_tables_c.setText(_translate("Dialog", "..."))
        self.label_2.setText(_translate("Dialog", "Hierarchy:"))
        self.sil_c_label.setText(_translate("Dialog", "SIL C:"))
        self.pick_sil_c.setText(_translate("Dialog", "..."))
        self.label_3.setText(_translate("Dialog", "SunSpec Spreadsheet:"))
        self.tables_c_label.setText(_translate("Dialog", "Tables C:"))
        self.remove_smdx.setText(_translate("Dialog", "Remove"))
        self.pick_spreadsheet.setText(_translate("Dialog", "..."))
        self.pick_smdx.setText(_translate("Dialog", "..."))
        self.pick_spreadsheet_user.setText(_translate("Dialog", "..."))
        self.label.setText(_translate("Dialog", "CAN:"))
        self.sunspec_tables_c_label.setText(_translate("Dialog", "SunSpec Tables C:"))
        self.pick_tables_c.setText(_translate("Dialog", "..."))
        self.from_directory.setText(_translate("Dialog", "From directory..."))
        self.pick_hierarchy.setText(_translate("Dialog", "..."))
        self.label_4.setText(_translate("Dialog", "SunSpec User Spreadsheet:"))
        self.interface_c_label.setText(_translate("Dialog", "Interface C:"))
        self.pick_interface_c.setText(_translate("Dialog", "..."))
        self.sunspec_c_label.setText(_translate("Dialog", "SunSpec C:"))
        self.pick_can.setText(_translate("Dialog", "..."))
        self.smdx_label.setText(_translate("Dialog", "SunSpec SMDX:"))
        self.pick_sunspec_c.setText(_translate("Dialog", "..."))
        self.sunspec_bitfields_c_label.setText(_translate("Dialog", "SunSpec Bitfields C"))
        self.pick_sunspec_bitfields_c.setText(_translate("Dialog", "..."))



import functools
import os
import textwrap

import click.testing
import pycparser.c_ast
import pycparser.c_generator
import pytest

import epyqlib.attrsmodel
import epyqlib.pm.parametermodel

import epcpm.parameterstoc


def disabled_test_exploration():
    path = os.path.join(os.path.dirname(__file__), 'example_parameters.json')

    runner = click.testing.CliRunner()
    runner.isolated_filesystem()
    result = runner.invoke(
        epcpm.parameterstoc.cli,
        [
            '--parameters', path,
        ],
    )

    print(result.output)

    assert result.exit_code == 0


def test_pycparser_exploration_parse():
    sample = '''
    typedef int int16_t;
    typedef int uint16_t;

    enum EnumName
    {
        one=1,
        two=2
    };
    typedef enum EnumName EnumTypedefName;

    struct StructName
    {
        int16_t a;
        uint16_t b;
    };
    typedef struct StructName StructTypedefName;
    
    int16_t array[5];
    '''

    parser = pycparser.CParser()
    ast = parser.parse(sample)

    generator = pycparser.c_generator.CGenerator()

    generator.visit(ast)

    return ast


def test_pycparser_exploration_wrapped():
    top_level = []

    top_level.extend(epcpm.parameterstoc.enum(
        name='EnumName',
        enumerators=(
            ('a', 1),
            ('b', 2),
        ),
    ))

    top_level.extend(epcpm.parameterstoc.struct(
        name='StructName',
        member_decls=(
            epcpm.parameterstoc.Decl(
                type=epcpm.parameterstoc.Type(
                    name=name,
                    type=type,
                )
            )
            for type, name in (
                ('int16_t', 'a'),
                ('uint16_t', 'b'),
            )
        )
    ))

    ast = pycparser.c_ast.FileAST(top_level)

    generator = pycparser.c_generator.CGenerator()

    s = generator.visit(ast)
    assert s == textwrap.dedent('''\
    enum EnumName_e
    {
      a = 1,
      b = 2
    };
    typedef enum EnumName_e EnumName_et;
    struct StructName_s
    {
      int16_t a;
      uint16_t b;
    };
    typedef struct StructName_s StructName_t;
    ''')


def test_single_layer_group_to_c():
    group = epyqlib.pm.parametermodel.Group(
        name='Group Name',
    )

    children = [
        epyqlib.pm.parametermodel.Parameter(
            name='Parameter A',
            type_name='int16_t',
        ),
        epyqlib.pm.parametermodel.Parameter(
            name='Parameter B',
            type_name='int16_t',
        ),
        epyqlib.pm.parametermodel.Parameter(
            name='Parameter C',
            type_name='int16_t',
        ),
    ]

    for child in children:
        group.append_child(child)

    builder = epcpm.parameterstoc.builders.wrap(group)

    ast = pycparser.c_ast.FileAST(builder.definition())
    generator = pycparser.c_generator.CGenerator()
    s = generator.visit(ast)

    assert s == textwrap.dedent('''\
        struct GroupName_s
        {
          int16_t parameterA;
          int16_t parameterB;
          int16_t parameterC;
        };
        typedef struct GroupName_s GroupName_t;
        ''')


def test_nested_group_to_c():
    inner_inner_group = epyqlib.pm.parametermodel.Group(
        name='Inner Inner Group Name',
    )

    children = [
        epyqlib.pm.parametermodel.Parameter(
            name='Parameter F',
            type_name='int16_t',
        ),
        epyqlib.pm.parametermodel.Parameter(
            name='Parameter G',
            type_name='int16_t',
        ),
    ]

    for child in children:
        inner_inner_group.append_child(child)

    inner_group = epyqlib.pm.parametermodel.Group(
        name='Inner Group Name',
    )

    children = [
        epyqlib.pm.parametermodel.Parameter(
            name='Parameter D',
            type_name='int16_t',
        ),
        inner_inner_group,
        epyqlib.pm.parametermodel.Parameter(
            name='Parameter E',
            type_name='int16_t',
        ),
    ]

    for child in children:
        inner_group.append_child(child)

    outer_group = epyqlib.pm.parametermodel.Group(
        name='Outer Group Name',
    )

    children = [
        epyqlib.pm.parametermodel.Parameter(
            name='Parameter A',
            type_name='int16_t',
        ),
        inner_group,
        epyqlib.pm.parametermodel.Parameter(
            name='Parameter B',
            type_name='int16_t',
        ),
        epyqlib.pm.parametermodel.Parameter(
            name='Parameter C',
            type_name='int16_t',
        ),
    ]

    for child in children:
        outer_group.append_child(child)

    builder = epcpm.parameterstoc.builders.wrap(outer_group)

    ast = pycparser.c_ast.FileAST(builder.definition())
    generator = pycparser.c_generator.CGenerator()
    s = generator.visit(ast)

    assert s == textwrap.dedent('''\
        struct InnerInnerGroupName_s
        {
          int16_t parameterF;
          int16_t parameterG;
        };
        typedef struct InnerInnerGroupName_s InnerInnerGroupName_t;
        struct InnerGroupName_s
        {
          int16_t parameterD;
          InnerInnerGroupName_t innerInnerGroupName;
          int16_t parameterE;
        };
        typedef struct InnerGroupName_s InnerGroupName_t;
        struct OuterGroupName_s
        {
          int16_t parameterA;
          InnerGroupName_t innerGroupName;
          int16_t parameterB;
          int16_t parameterC;
        };
        typedef struct OuterGroupName_s OuterGroupName_t;
        ''')


def test_datalogger_a():
    data_logger = epyqlib.pm.parametermodel.Group(name='Data Logger')

    chunks_array = epyqlib.pm.parametermodel.Array(name='Chunks')
    data_logger.append_child(chunks_array)

    chunk = epyqlib.pm.parametermodel.Group(type_name='Chunk')
    chunks_array.append_child(chunk)
    chunks_array.length = 4
    chunks_array.children[0].name = 'First'
    chunks_array.children[1].name = 'Second'
    chunks_array.children[2].name = 'Third'
    chunks_array.children[3].name = 'Fourth'

    address = epyqlib.pm.parametermodel.Parameter(
        name='Address',
        default=0,
        type_name='int16_t',
    )
    chunk.append_child(address)
    bytes_ = epyqlib.pm.parametermodel.Parameter(
        name='Bytes',
        default=0,
        type_name='int16_t',
    )
    chunk.append_child(bytes_)

    post_trigger_duration = epyqlib.pm.parametermodel.Parameter(
        name='Post Trigger Duration',
        default=500,
        type_name='int16_t',
    )
    data_logger.append_child(post_trigger_duration)

    group = epyqlib.pm.parametermodel.Group(name='Group')
    data_logger.append_child(group)

    param = epyqlib.pm.parametermodel.Parameter(
        name='Param',
        type_name='int16_t',
    )
    group.append_child(param)

    builder = epcpm.parameterstoc.builders.wrap(data_logger)

    ast = pycparser.c_ast.FileAST(builder.definition())
    generator = pycparser.c_generator.CGenerator()
    s = generator.visit(ast)

    assert s == textwrap.dedent('''\
    struct Chunk_s
    {
      int16_t address;
      int16_t bytes;
    };
    typedef struct Chunk_s Chunk_t;
    enum Chunks_e
    {
      Chunks_First = 0,
      Chunks_Second = 1,
      Chunks_Third = 2,
      Chunks_Fourth = 3,
      Chunks_Count = 4
    };
    typedef enum Chunks_e Chunks_et;
    typedef Chunk_t Chunks_t[Chunks_Count];
    struct Group_s
    {
      int16_t param;
    };
    typedef struct Group_s Group_t;
    struct DataLogger_s
    {
      Chunks_t chunks;
      int16_t postTriggerDuration;
      Group_t group;
    };
    typedef struct DataLogger_s DataLogger_t;
    ''')

data_logger_structures = '''\
typedef struct
{
    void*	address;
    size_t	bytes;
} DataLogger_Chunk;

#define DATALOGGER_CHUNK_DEFAULTS(...) \
{ \
    .address = 0, \
    .bytes = 0, \
    __VA_ARGS__ \
}

typedef DataLogger_Chunk DataLogger_Chunks[dataLoggerChunkCount];

typedef struct
{
    DataLogger_Chunks chunks;
    uint16_t postTriggerDuration_ms;
} DataLogger_Params;

#define DATALOGGER_PARAMS_DEFAULTS(...) \
{ \
    .chunks = {[0 ... dataLoggerChunkCount-1] = DATALOGGER_CHUNK_DEFAULTS()}, \
    .postTriggerDuration_ms = 500, \
    __VA_ARGS__ \
}
'''


def test_basic_parameter_array():
    array = epyqlib.pm.parametermodel.Array(name='Array Name')
    parameter = epyqlib.pm.parametermodel.Parameter(
        name='Parameter Name',
        type_name='int16_t',
    )

    array.append_child(parameter)

    n = 5

    array.length = n

    builder = epcpm.parameterstoc.builders.wrap(array)

    ast = pycparser.c_ast.FileAST(builder.definition())
    generator = pycparser.c_generator.CGenerator()
    s = generator.visit(ast)

    assert s == textwrap.dedent('''\
    enum ArrayName_e
    {
      ArrayName_ParameterName = 0,
      ArrayName__2 = 1,
      ArrayName__3 = 2,
      ArrayName__4 = 3,
      ArrayName__5 = 4,
      ArrayName_Count = 5
    };
    typedef enum ArrayName_e ArrayName_et;
    typedef int16_t ArrayName_t[ArrayName_Count];
    ''')


def test_grouped_parameter_array():
    group = epyqlib.pm.parametermodel.Group(name='Group Name')
    array = epyqlib.pm.parametermodel.Array(name='Array Name')
    parameter = epyqlib.pm.parametermodel.Parameter(
        name='Parameter Name',
        type_name='int16_t',
    )

    group.append_child(array)
    array.append_child(parameter)

    n = 5

    array.length = n

    builder = epcpm.parameterstoc.builders.wrap(group)

    ast = pycparser.c_ast.FileAST(builder.definition())
    generator = pycparser.c_generator.CGenerator()
    s = generator.visit(ast)

    assert s == textwrap.dedent('''\
    enum ArrayName_e
    {
      ArrayName_ParameterName = 0,
      ArrayName__2 = 1,
      ArrayName__3 = 2,
      ArrayName__4 = 3,
      ArrayName__5 = 4,
      ArrayName_Count = 5
    };
    typedef enum ArrayName_e ArrayName_et;
    typedef int16_t ArrayName_t[ArrayName_Count];
    struct GroupName_s
    {
      ArrayName_t arrayName;
    };
    typedef struct GroupName_s GroupName_t;
    ''')


def test_grouped_parameter_array_no_enum():
    array = epyqlib.pm.parametermodel.Array(name='Array Name')
    array.named_enumerators = False
    parameter = epyqlib.pm.parametermodel.Parameter(
        name='Parameter Name',
        type_name='int16_t',
    )

    array.append_child(parameter)

    n = 5

    array.length = n

    builder = epcpm.parameterstoc.builders.wrap(array)

    ast = pycparser.c_ast.FileAST(builder.definition())
    generator = pycparser.c_generator.CGenerator()
    s = generator.visit(ast)

    assert s == textwrap.dedent('''\
    enum ArrayName_e
    {
      ArrayName_Count = 5
    };
    typedef enum ArrayName_e ArrayName_et;
    typedef int16_t ArrayName_t[ArrayName_Count];
    ''')


def test_line_monitor_params():
    line_monitoring = epyqlib.pm.parametermodel.Group(
        name='Line Monitoring',
    )

    frequency_limits = epyqlib.pm.parametermodel.Array(
        name='Frequency Limits',
    )
    line_monitoring.append_child(frequency_limits)

    frequency_limit = epyqlib.pm.parametermodel.Group(
        name='First',
        type_name='Frequency Limit',
    )
    frequency_limits.append_child(frequency_limit)

    frequency = epyqlib.pm.parametermodel.Parameter(
        name='Frequency',
        type_name='_iq',
    )
    frequency_limit.append_child(frequency)

    clearing_time = epyqlib.pm.parametermodel.Parameter(
        name='Clearing Time',
        type_name='_iq',
    )
    frequency_limit.append_child(clearing_time)

    frequency_limits.length = 4
    frequency_limits.children[1].name = 'Second'
    frequency_limits.children[2].name = 'Third'
    frequency_limits.children[3].name = 'Fourth'

    builder = epcpm.parameterstoc.builders.wrap(line_monitoring)

    ast = pycparser.c_ast.FileAST(builder.definition())
    generator = pycparser.c_generator.CGenerator()
    s = generator.visit(ast)

    assert s == textwrap.dedent('''\
    struct FrequencyLimit_s
    {
      _iq frequency;
      _iq clearingTime;
    };
    typedef struct FrequencyLimit_s FrequencyLimit_t;
    enum FrequencyLimits_e
    {
      FrequencyLimits_First = 0,
      FrequencyLimits_Second = 1,
      FrequencyLimits_Third = 2,
      FrequencyLimits_Fourth = 3,
      FrequencyLimits_Count = 4
    };
    typedef enum FrequencyLimits_e FrequencyLimits_et;
    typedef FrequencyLimit_t FrequencyLimits_t[FrequencyLimits_Count];
    struct LineMonitoring_s
    {
      FrequencyLimits_t frequencyLimits;
    };
    typedef struct LineMonitoring_s LineMonitoring_t;
    ''')


def test_root():
    root = epyqlib.pm.parametermodel.Root()
    group = epyqlib.pm.parametermodel.Group(name='Group')
    p1 = epyqlib.pm.parametermodel.Parameter(name='red', type_name='RedType')
    p2 = epyqlib.pm.parametermodel.Parameter(name='blue', type_name='BlueType')

    root.append_child(group)
    group.append_child(p1)
    root.append_child(p2)

    builder = epcpm.parameterstoc.builders.wrap(root)

    ast = pycparser.c_ast.FileAST(builder.definition())
    generator = pycparser.c_generator.CGenerator()
    s = generator.visit(ast)

    assert s == textwrap.dedent('''\
    struct Group_s
    {
      RedType red;
    };
    typedef struct Group_s Group_t;
    ''')

    ast = pycparser.c_ast.FileAST(builder.instantiation())
    s = generator.visit(ast)

    assert s == textwrap.dedent('''\
    Group_t group;
    BlueType blue;
    ''')



import collections
import contextlib
import functools
import io
import logging
import os
import pathlib

import attr
import pycparser.c_ast
import pycparser.c_generator
from PyQt5 import QtCore, QtGui, QtWidgets

import epyqlib.attrsmodel
import epyqlib.checkresultmodel
import epyqlib.pm.parametermodel
import epyqlib.pm.valuesetmodel
import epyqlib.utils.qt

import epcpm.canmodel
import epcpm.cantosym
import epcpm.importexport
import epcpm.importexportdialog
import epcpm.parameterstoc
import epcpm.parameterstohierarchy
import epcpm.project
import epcpm.smdxtosunspec
import epcpm.sunspecmodel
import epcpm.sunspectoxlsx
import epcpm.symtoproject

import epcpm.mainwindow_ui


# See file COPYING in this source tree
__copyright__ = 'Copyright 2019, EPC Power Corp.'
__license__ = 'GPLv2+'


@attr.s
class ModelView:
    view = attr.ib()
    # filename = attr.ib()
    types = attr.ib()
    extras = attr.ib(default=attr.Factory(collections.OrderedDict))
    model = attr.ib(default=None)
    proxy = attr.ib(default=None)
    selection = attr.ib(default=None)


class Window:
    def __init__(self, title, icon_path):
        logging.debug('Working directory: {}'.format(os.getcwd()))

        self.main_window = QtWidgets.QMainWindow()
        self.ui = epcpm.mainwindow_ui.Ui_MainWindow()
        self.ui.setupUi(self.main_window)

        self.ui.action_new_project.triggered.connect(
            lambda _: self.open_project(),
        )
        self.ui.action_open_project.triggered.connect(
            lambda _: self.open_project_from_dialog(),
        )
        self.ui.action_save_project.triggered.connect(
            lambda _: self.save_project(),
        )
        self.ui.action_save_as_project.triggered.connect(
            lambda _: self.save_as_project(),
        )

        self.ui.action_new_value_set.triggered.connect(
            lambda _: self.new_value_set(),
        )
        self.ui.action_open_value_set.triggered.connect(
            lambda _: self.open_value_set_from_dialog(),
        )
        self.ui.action_save_value_set.triggered.connect(
            lambda _: self.save_value_set(),
        )
        self.ui.action_save_as_value_set.triggered.connect(
            lambda _: self.save_as_value_set(),
        )
        self.ui.action_check.triggered.connect(
            lambda _: self.check(),
        )

        self.ui.action_import_sym.triggered.connect(self.import_sym)
        self.ui.action_export_sym.triggered.connect(self.generate_symbol_file)
        self.ui.action_import_smdx.triggered.connect(self.import_smdx)
        self.ui.action_full_import.triggered.connect(self.full_import)
        self.ui.action_full_export.triggered.connect(self.full_export)
        self.ui.action_full_export_first_time.triggered.connect(
            self.full_export_first_time,
        )

        self.ui.action_about.triggered.connect(self.about_dialog)

        self.main_window.setWindowTitle(title)

        logging.debug('Loading icon from: {}'.format(icon_path))
        self.main_window.setWindowIcon(QtGui.QIcon(str(pathlib.Path(
            pathlib.Path(__file__).parents[0],
            icon_path,
        ))))

        self.project_filters = [
            ('Parameter Project', ['pmp']),
            ('All Files', ['*'])
        ]

        self.data_filters = [
            ('JSON', ['json']),
            ('All Files', ['*'])
        ]
        self.hierarchy_filters = self.data_filters

        self.can_filters = [
            ('CAN Symbols', ['sym']),
            ('All Files', ['*'])
        ]

        self.smdx_filters = [
            ('SunSpec SMDX', ['xml']),
            ('All Files', ['*'])
        ]

        self.view_models = {}

        self.uuid_notifiers = {
            'can': epcpm.canmodel.ReferencedUuidNotifier(),
            'sunspec': epcpm.sunspecmodel.ReferencedUuidNotifier(),
            'check_result': epyqlib.checkresultmodel.ReferencedUuidNotifier(),
        }
        self.uuid_notifiers['can'].changed.connect(self.can_uuid_changed)
        self.uuid_notifiers['sunspec'].changed.connect(self.can_uuid_changed)
        self.uuid_notifiers['check_result'].changed.connect(
            self.check_result_uuid_changed,
        )

        self.project = None
        self.value_set = None
        self.check_result = None

        self.set_title()

        search_boxes = (
            (
                self.ui.parameter_view,
                self.ui.parameter_search_box,
                epyqlib.pm.parametermodel.columns.index_of('Name'),
            ),
            (
                self.ui.can_view,
                self.ui.can_search_box,
                epcpm.canmodel.columns.index_of('Name'),
            ),
            (
                self.ui.sunspec_view,
                self.ui.sunspec_search_box,
                epcpm.sunspecmodel.columns.index_of('Name'),
            ),
            (
                self.ui.value_set_view,
                self.ui.value_set_search_box,
                epyqlib.pm.valuesetmodel.columns.index_of('Name'),
            ),
        )

        for view, search_box, column in search_boxes:
            search_box.connect_to_view(view=view, column=column)

        self.ui.check_result_view.setEditTriggers(
            QtWidgets.QAbstractItemView.NoEditTriggers,
        ),

    def set_title(self, detail=None):
        title = 'Parameter Manager v{}'.format(epcpm.__version__)

        if detail is not None:
            title = ' - '.join((title, detail))

        self.main_window.setWindowTitle(title)

    def set_model(self, name, view_model):
        self.view_models[name] = view_model

        view_model.proxy = epyqlib.utils.qt.PySortFilterProxyModel(
            filter_column=None,
        )
        view_model.proxy.setSortCaseSensitivity(QtCore.Qt.CaseInsensitive)
        view_model.proxy.setSourceModel(view_model.model.model)
        view_model.view.setModel(view_model.proxy)
        view_model.view.setItemDelegate(epyqlib.attrsmodel.create_delegate())

        notifier = self.uuid_notifiers.get(name)
        if notifier is not None:
            notifier.set_view(view_model.view)

        view_model.selection = view_model.view.selectionModel()

        with contextlib.suppress(TypeError):
            view_model.selection.selectionChanged.disconnect(
                self.selection_changed,
            )

        view_model.selection.selectionChanged.connect(
            self.selection_changed,
        )

    def open_project_from_dialog(self):
        filename = epyqlib.utils.qt.file_dialog(
            self.project_filters,
            parent=self.main_window,
        )

        if filename is None:
            return

        self.open_project(filename=filename)

    def import_sym(self):
        sym_path = epyqlib.utils.qt.file_dialog(
            filters=self.can_filters,
            parent=self.main_window,
        )

        if sym_path is None:
            return

        hierarchy_path = epyqlib.utils.qt.file_dialog(
            filters=self.hierarchy_filters,
            parent=self.main_window,
            caption='Open Parameter Hierarchy',
        )

        if hierarchy_path is None:
            return

        with open(sym_path, 'rb') as sym, open(hierarchy_path) as hierarchy:
            parameters_root, can_root, sunspec_root = (
                epcpm.symtoproject.load_can_file(
                    can_file=sym,
                    file_type=str(pathlib.Path(sym.name).suffix[1:]),
                    parameter_hierarchy_file=hierarchy,
                )
            )

        project = epcpm.project.Project()

        project.models.parameters = epyqlib.attrsmodel.Model(
            root=parameters_root,
            columns=epyqlib.pm.parametermodel.columns,
        )
        project.models.can = epyqlib.attrsmodel.Model(
            root=can_root,
            columns=epcpm.canmodel.columns,
        )
        project.models.sunspec = epyqlib.attrsmodel.Model(
            root=sunspec_root,
            columns=epcpm.sunspecmodel.columns,
        )

        epcpm.project._post_load(project)

        # TODO: backmatching
        epcpm.symtoproject.go_add_tables(
            parameters_root=project.models.parameters.root,
            can_root=project.models.can.root,
        )

        self.open_project(project=project)

    def import_smdx(self):
        selected = epyqlib.utils.qt.file_dialog(
            filters=self.smdx_filters,
            parent=self.main_window,
            path_factory=pathlib.Path,
            multiple=True,
        )

        for smdx_path in selected:
            id = int(
                smdx_path.name.partition('smdx_')[2].rpartition('.xml')[0],
            )

            model = epcpm.smdxtosunspec.import_model(
                model_id=id,
                parameter_model=self.view_models['parameters'].model,
                paths=[smdx_path.parent],
            )

            self.view_models['sunspec'].model.root.append_child(model)

    def full_import(self):
        dialog = epcpm.importexportdialog.import_dialog()
        if dialog.exec() != QtWidgets.QDialog.DialogCode.Accepted:
            return

        project = epcpm.importexport.full_import(paths=dialog.paths_result)

        self.open_project(project=project)

        QtWidgets.QMessageBox.information(
            self.main_window,
            'Import Complete',
            'Import complete.',
        )

    def full_export(self):
        return self._full_export(first_time=False)

    def full_export_first_time(self):
        return self._full_export(first_time=True)

    def _full_export(self, first_time):
        dialog = epcpm.importexportdialog.export_dialog()
        if dialog.exec() != QtWidgets.QDialog.DialogCode.Accepted:
            return

        paths = dialog.paths_result

        epcpm.importexport.full_export(
            project=self.project,
            paths=paths,
            target_directory=dialog.directory,
            first_time=first_time,
            include_uuid_in_item=True,
        )

        QtWidgets.QMessageBox.information(
            self.main_window,
            'Export Complete',
            'Export complete.',
        )

    def open_project(self, filename=None, project=None):
        if project is not None:
            self.project = project
        else:
            if filename is None:
                self.project = epcpm.project.create_blank()
            else:
                self.project = epcpm.project.loadp(filename)

        model_views = epcpm.project.Models()

        model_views.parameters = ModelView(
            view=self.ui.parameter_view,
            types=epyqlib.pm.parametermodel.types,
            extras=collections.OrderedDict((
                ('Generate code...', self.generate_code),
            )),
        )

        model_views.can = ModelView(
            view=self.ui.can_view,
            types=epcpm.canmodel.types,
        )

        model_views.sunspec = ModelView(
            view=self.ui.sunspec_view,
            types=epcpm.sunspecmodel.types,
        )

        for notifier in self.uuid_notifiers.values():
            notifier.disconnect_view()

        i = zip(model_views.items(), self.project.models.values())
        for (name, model_view), model in i:
            view = model_view.view

            view.setAutoScroll(True)
            view.setAutoScrollMargin(32)
            view.setDropIndicatorShown(True)
            view.setDragEnabled(True)
            view.setAcceptDrops(True)

            if model is None:
                model_view.model = epyqlib.attrsmodel.Model(
                    root=model_view.root_factory(),
                    columns=model_view.columns,
                )
            else:
                model_view.model = model

            model.update_nodes()

            self.set_model(name=name, view_model=model_view)
            view.collapseAll()
            column_count = (
                model_view.model.model.columnCount(QtCore.QModelIndex())
            )
            for i in range(column_count):
                view.resizeColumnToContents(i)

            header = view.header()
            header.setSectionResizeMode(
                0,
                QtWidgets.QHeaderView.ResizeToContents,
            )

            view.setContextMenuPolicy(
                QtCore.Qt.CustomContextMenu)
            m = functools.partial(
                self.context_menu,
                view_model=model_view
            )

            with contextlib.suppress(TypeError):
                view.customContextMenuRequested.disconnect()

            view.customContextMenuRequested.connect(m)

        self.uuid_notifiers['can'].set_view(self.ui.can_view)
        self.uuid_notifiers['sunspec'].set_view(self.ui.sunspec_view)
        self.uuid_notifiers['check_result'].set_view(self.ui.check_result_view)

        return

    def save_project(self):
        self.project.save(parent=self.main_window)

    def save_as_project(self):
        project = attr.evolve(self.project)
        project.filename = None
        # TODO: this is still going to mutate the same object as the
        #       original project is referencing
        project.paths.set_all(None)

        project.save(parent=self.main_window)
        self.project = project

    def new_value_set(self):
        parameters = self.view_models.get('parameters')
        if parameters is not None:
            parameters = parameters.model

        value_set = epyqlib.pm.valuesetmodel.create_blank(
            parameter_model=parameters,
        )

        human_names = QtWidgets.QMessageBox.question(
            self.main_window,
            'Humanized Names',
            'Use humanized names in new value set?',
            buttons=(
                QtWidgets.QMessageBox.Yes
                | QtWidgets.QMessageBox.No
            ),
            defaultButton=QtWidgets.QMessageBox.Yes,
        ) == QtWidgets.QMessageBox.Yes

        only_parameters = QtWidgets.QMessageBox.question(
            self.main_window,
            'Only Parameters',
            'Only include parameters group in new value set?',
            buttons=(
                QtWidgets.QMessageBox.Yes
                | QtWidgets.QMessageBox.No
            ),
            defaultButton=QtWidgets.QMessageBox.No,
        ) == QtWidgets.QMessageBox.Yes

        calculate_unspecified_min_max = QtWidgets.QMessageBox.question(
            self.main_window,
            'Calculate Unspecified',
            (
                'Calculate unspecified minimum And maximum values from '
                'symbol ranges?'
            ),
            buttons=(
                QtWidgets.QMessageBox.Yes
                | QtWidgets.QMessageBox.No
            ),
            defaultButton=QtWidgets.QMessageBox.No,
        ) == QtWidgets.QMessageBox.Yes

        base_node = None
        if only_parameters:
            base_node, = [
                node
                for node in value_set.parameter_model.root.children
                if node.name == 'Parameters'
            ]

        epyqlib.pm.valuesetmodel.copy_parameter_data(
            value_set=value_set,
            human_names=human_names,
            base_node=base_node,
            calculate_unspecified_min_max=calculate_unspecified_min_max,
            can_root=self.view_models.get('can').model.root,
        )

        self.set_active_value_set(value_set)

    def set_active_value_set(self, value_set):
        self.value_set = value_set

        view = self.ui.value_set_view
        model = self.value_set.model

        model_view = ModelView(
            view=view,
            model=model,
            types=epyqlib.pm.valuesetmodel.types,
        )

        view.setSelectionBehavior(view.SelectRows)
        view.setSelectionMode(view.SingleSelection)
        view.setDropIndicatorShown(True)
        view.setDragEnabled(True)
        view.setAcceptDrops(True)

        self.set_model(name='value_set', view_model=model_view)
        view.expandAll()

    def open_value_set_from_dialog(self):
        path = epyqlib.utils.qt.file_dialog(
            attr.fields(epyqlib.pm.valuesetmodel.ValueSet).filters.default,
            parent=self.main_window,
        )

        if path is None:
            return

        value_set = epyqlib.pm.valuesetmodel.loadp(path)

        self.set_active_value_set(value_set)

    def save_value_set(self):
        self.value_set.save(parent=self.main_window)

    def save_as_value_set(self):
        value_set = attr.evolve(self.value_set)
        value_set.path = None

        value_set.save(parent=self.main_window)
        self.value_set = value_set

    def check(self):
        root = epyqlib.checkresultmodel.Root()

        for name, model in self.project.models.items():
            node = epyqlib.checkresultmodel.Node.build(
                name=name,
                node=model.root,
            )
            model.root.check_and_append(
                parent=node,
                models=self.project.models,
            )
            root.append_child(node)

        # self.check_view = QtWidgets.QTreeView()
        # self.check_model = epyqlib.attrsmodel.Model(
        #     root=root,
        #     columns=epyqlib.checkresultmodel.columns,
        # )

        # self.check_view.setModel(self.check_model.model)

        # self.check_view.show()

        drop_source_model_names = [
            'parameters',
            'can',
            'sunspec',
        ]

        model = epyqlib.attrsmodel.Model(
            root=root,
            columns=epyqlib.checkresultmodel.columns,
            drop_sources=[
                self.view_models[name].model
                for name in drop_source_model_names
            ],
        )
        self.set_active_check_result(model)

    def set_active_check_result(self, check_result):
        self.check_result = check_result

        view = self.ui.check_result_view

        model_view = ModelView(
            view=view,
            model=self.check_result,
            types=epyqlib.checkresultmodel.types,
        )

        view.setSelectionBehavior(view.SelectRows)
        view.setSelectionMode(view.SingleSelection)
        view.setDropIndicatorShown(True)
        view.setDragEnabled(False)
        view.setAcceptDrops(False)

        self.set_model(name='check_result', view_model=model_view)
        view.expandAll()

    def context_menu(self, position, view_model):
        view = view_model.view
        view_index = view.indexAt(position)
        index_is_valid = view_index.isValid()
        index = view.model().mapToSource(view_index)

        model = view_model.model
        node = model.node_from_index(index)

        menu = QtWidgets.QMenu(parent=view)

        node_type = type(node)
        type_label = menu.addAction(
            '{}.{}'.format(node_type.__module__, node_type.__name__),
        )
        type_label.setDisabled(True)

        menu.addSeparator()

        add_submenu = menu.addMenu('Add')

        addable_types = node.addable_types()
        actions = {
            add_submenu.addAction('Add {}'.format(name)): t
            for name, t in addable_types.items()
        }
        add_submenu.setDisabled(len(actions) == 0)

        update = menu.addAction('Update')
        update.setEnabled(hasattr(node, 'update'))

        menu.addSeparator()

        delete = menu.addAction('&Delete')
        delete.setEnabled(node.can_delete())

        menu.addSeparator()

        extra_actions = {
            menu.addAction(name): function
            for name, function in view_model.extras.items()
        }

        if len(extra_actions) > 0:
            menu.addSeparator()

        expand_tree = menu.addAction('Expand Tree')
        expand_tree.setEnabled(index_is_valid)
        expand_all = menu.addAction('Expand All')
        menu.addSeparator()
        collapse_tree = menu.addAction('Collapse Tree')
        collapse_tree.setEnabled(index_is_valid)
        collapse_all = menu.addAction('Collapse All')

        action = menu.exec(
            view.viewport().mapToGlobal(position)
        )

        if action is not None:
            extra = extra_actions.get(action)
            if extra is not None:
                extra(node)
            elif action is delete:
                node.tree_parent.remove_child(child=node)
            elif action is update:
                node.update()
            elif action is expand_tree:
                epyqlib.utils.qt.set_expanded_tree(
                    view=view,
                    index=view_index,
                    expanded=True,
                )
            elif action is expand_all:
                view.expandAll()
            elif action is collapse_tree:
                epyqlib.utils.qt.set_expanded_tree(
                    view=view,
                    index=view_index,
                    expanded=False,
                )
            elif action is collapse_all:
                view.collapseAll()
            else:
                new_node = actions[action]()
                node.append_child(new_node)
                new_index = model.index_from_node(new_node)
                view_index = epyqlib.utils.qt.resolve_index_from_model(
                    model=model.model,
                    view=view,
                    index=new_index,
                )
                view.setCurrentIndex(view_index)
                view.edit(view_index)

    def generate_code(self, node):
        builder = epcpm.parameterstoc.builders.wrap(node)
        ast = pycparser.c_ast.FileAST(builder.definition())
        generator = pycparser.c_generator.CGenerator()
        s = generator.visit(ast)
        epyqlib.utils.qt.dialog(
            parent=self.main_window,
            message=s,
            modal=False,
        )

    def generate_symbol_file(self):
        finder = self.view_models['can'].model.node_from_uuid
        access_levels = self.project.models.parameters.list_selection_roots[
            'access level'
        ]
        builder = epcpm.cantosym.builders.wrap(
            wrapped=self.view_models['can'].model.root,
            access_levels=access_levels,
            parameter_uuid_finder=finder,
            parameter_model=self.view_models['parameters'].model,
        )

        epyqlib.utils.qt.dialog(
            parent=self.main_window,
            message=builder.gen(),
            modal=False,
            save_filters=(
                ('CAN Symbols', ['sym']),
                ('All Files', ['*'])
            ),
            save_caption='Save CAN Symbols',
        )

    def selection_changed(self, selected, deselected):
        pass

    # TODO: CAMPid 0795409054128050124650546086
    def can_uuid_changed(self, uuid):
        view_model = self.view_models['parameters']
        model = view_model.model
        view = view_model.view

        try:
            node = model.node_from_uuid(uuid)
        except epyqlib.attrsmodel.NotFoundError:
            return

        index = model.index_from_node(node)
        index = epyqlib.utils.qt.resolve_index_from_model(
            model=model.model,
            view=view,
            index=index,
        )

        view.setCurrentIndex(index)
        view.selectionModel().select(
            index,
            (
                QtCore.QItemSelectionModel.ClearAndSelect
                | QtCore.QItemSelectionModel.Rows
            ),
        )

    # TODO: CAMPid 0795409054128050124650546086
    def check_result_uuid_changed(self, uuid):
        view_model = self.view_models['check_result']
        model = view_model.model
        view = view_model.view

        try:
            node = model.node_from_uuid(uuid)
        except epyqlib.attrsmodel.NotFoundError:
            return

        index = model.index_from_node(node)
        for view_model in self.view_models.values():
            if index.model() is not view_model.model.model:
                continue

            model = view_model.model
            view = view_model.view

            index = epyqlib.utils.qt.resolve_index_from_model(
                model=model.model,
                view=view,
                index=index,
            )

            view.setCurrentIndex(index)
            view.selectionModel().select(
                index,
                QtCore.QItemSelectionModel.ClearAndSelect,
            )

    def about_dialog(self):
        message = [
            __copyright__,
            __license__,
            f'Version Tag: {epcpm.__version_tag__}',
            f'Commit SHA: {epcpm.__sha__}',
            f'Build Tag: {epcpm.__build_tag__}',
        ]

        message = '\n'.join(message)

        epyqlib.utils.qt.dialog(
            parent=self.main_window,
            title='About',
            message=message,
        )



import collections
import json
import pathlib

import attr
import graham
import pytest

import epyqlib.attrsmodel
import epyqlib.pm.parametermodel
import epyqlib.tests.pm.test_parametermodel
import epyqlib.tests.test_attrsmodel

import epcpm.canmodel
import epcpm.project

# See file COPYING in this source tree
__copyright__ = 'Copyright 2017, EPC Power Corp.'
__license__ = 'GPLv2+'


here = pathlib.Path(__file__).parent

with open(here/'project'/'can.json') as f:
    serialized_sample = f.read()


@attr.s
class SampleModelFromFile:
    root = attr.ib()
    parameter_root = attr.ib()
    model = attr.ib()
    parameter_model = attr.ib()
    table = attr.ib()
    parameter_table = attr.ib()
    project = attr.ib()

    @classmethod
    def build(cls):
        project = epcpm.project.loadp(here / 'project' / 'project.pmp')
        can_table, = project.models.can.root.nodes_by_attribute(
            attribute_value='First Table',
            attribute_name='name',
        )
        parameter_table = project.models.parameters.node_from_uuid(
            can_table.table_uuid,
        )

        sample_model = cls(
            root=project.models.can.root,
            parameter_root=project.models.parameters.root,
            model=project.models.can,
            parameter_model=project.models.parameters,
            table=can_table,
            parameter_table=parameter_table,
            project=project,
        )

        return sample_model


@pytest.fixture
def sample():
    sample_model = SampleModelFromFile.build()
    sample_model.parameter_table.update()
    sample_model.table.update()

    return sample_model


def test_hex_field():
    @graham.schemify(tag='test')
    @attr.s(hash=False)
    class Test:
        field = attr.ib(
            metadata=graham.create_metadata(
                field=epcpm.canmodel.HexadecimalIntegerField(),
            ),
        )
        none = attr.ib(
            default=None,
            metadata=graham.create_metadata(
                field=epcpm.canmodel.HexadecimalIntegerField(
                    allow_none=True,
                ),
            ),
        )

    t = Test(field=0x123)

    serialized = '{"_type": "test", "field": "0x123", "none": null}'

    assert graham.dumps(t).data == serialized
    assert graham.schema(Test).loads(serialized).data == t


TestAttrsModel = epyqlib.attrsmodel.build_tests(
    types=epcpm.canmodel.types,
    root_type=epcpm.canmodel.Root,
    columns=epcpm.canmodel.columns,
)


@attr.s
class SampleModel:
    root = attr.ib(
        factory=lambda: epcpm.canmodel.Root(
            uuid='b92665a4-6deb-4faf-8747-3aa20cf0bcf2',
        ),
    )
    model = attr.ib(default=None)
    parameters_message = attr.ib(default=None)
    parameters_multiplexer = attr.ib(default=None)

    @classmethod
    def build(cls):
        sample_model = cls()
        sample_model.model = epyqlib.attrsmodel.Model(
            root=sample_model.root,
            columns=epcpm.canmodel.columns,
        )

        return sample_model

    def fill(self):
        self.parameters_message = epcpm.canmodel.MultiplexedMessage(
            uuid='e5228583-1e7f-4a8a-a529-bbd84b2d0fca',
        )
        self.root.append_child(self.parameters_message)

        self.parameters_multiplexer = epcpm.canmodel.Multiplexer(
            uuid='70a1ab55-09eb-4617-b6a4-19ec821e7dfe',
        )
        self.parameters_message.append_child(self.parameters_multiplexer)

        self.model.update_nodes()


# TODO: CAMPid 094329054780541680163054608431067542971349
def count_types(sequence):
    counts = collections.defaultdict(int)

    for element in sequence:
        counts[type(element)] += 1

    return counts


def test_table_update_unlinked():
    expected_counts = {
        epcpm.canmodel.Signal: 2,
        epcpm.canmodel.Multiplexer: 24,
    }

    project = epcpm.project.loadp(here/'project'/'project.pmp')
    can_table, = project.models.can.root.nodes_by_attribute(
        attribute_value='First Table',
        attribute_name='name',
    )
    parameter_table = project.models.parameters.node_from_uuid(
        can_table.table_uuid,
    )
    parameter_table.update()

    assert count_types(can_table.children) == expected_counts

    can_table.update()
    assert count_types(can_table.children) == expected_counts

    can_table.table_uuid = None
    can_table.update()

    assert count_types(can_table.children) == {}


def test_table_update_same_uuid(sample):
    # TODO: CAMPid 9784566547216435136479765479163496731
    def collect(node):
        def collect(node, payload):
            payload[node.uuid] = node.name

        results = {}

        node.traverse(
            call_this=collect,
            internal_nodes=True,
            payload=results,
        )

        return results

    sample.table.update()

    original = collect(sample.table)

    sample.table.update()

    after_update = collect(sample.table)

    assert after_update == original


def test_sample_dumps_consistently(sample):
    dumped = graham.dumps(sample.root, indent=4)

    def load(s):
        return json.loads(s, object_pairs_hook=collections.OrderedDict)

    print()
    print(dumped.data)

    assert load(dumped.data) == load(serialized_sample)


def test_add_enumerator_update_table(sample):
    enumeration, = sample.parameter_root.nodes_by_attribute(
        attribute_value='Enumeration Two',
        attribute_name='name',
    )

    enumeration.append_child(epyqlib.pm.parametermodel.Enumerator(
        name='ET_New',
        value=42,
        uuid='173ba72c-bf2a-42a1-aab6-3e5fc49f10e7',
    ))

    sample.parameter_table.update()
    sample.table.update()


TestAttrsModel = epyqlib.attrsmodel.build_tests(
    types=epcpm.canmodel.types,
    root_type=epcpm.canmodel.Root,
    columns=epcpm.canmodel.columns,
)
